@@ -9,6 +9,7 @@ import warnings
 
 import numpy as np
 import torch
+from ray import tune
 from torch.optim.lr_scheduler import ReduceLROnPlateau
 
 from pytorch_widedeep.metrics import MultipleMetrics
@@ -159,7 +160,11 @@ class History(Callback):
         for k, v in logs.items():
             if isinstance(v, np.ndarray):
                 v = v.tolist()
-            self.trainer.history.setdefault(k, []).append(v)
+            if isinstance(v, list) and len(v) > 1:
+                for i in range(len(v)):
+                    self.trainer.history.setdefault(k + "_" + str(i), []).append(v[i])
+            else:
+                self.trainer.history.setdefault(k, []).append(v)
 
 
 class LRShedulerCallback(Callback):
@@ -333,12 +338,13 @@ class ModelCheckpoint(Callback):
 
     Parameters
     ----------
-    filepath: str
+    filepath: str, default=None
         Full path to save the output weights. It must contain only the root of
         the filenames. Epoch number and ``.pt`` extension (for pytorch) will
         be added. e.g. ``filepath="path/to/output_weights/weights_out"`` And
         the saved files in that directory will be named: ``weights_out_1.pt,
         weights_out_2.pt, ...``
+        If set to None the class just report best metric and best_epoch.
     monitor: str, default="loss"
         quantity to monitor. Typically 'val_loss' or metric name (e.g. 'val_acc')
     verbose:int, default=0
@@ -357,6 +363,15 @@ class ModelCheckpoint(Callback):
         Interval (number of epochs) between checkpoints.
     max_save: int, default=-1
         Maximum number of outputs to save. If -1 will save all outputs
+    wb: obj, default=None
+        Weights&Biases API interface to report single best result usable for comparisson of multiple
+        paramater combinations by e.g. parallel coordinates:
+        https://docs.wandb.ai/ref/app/features/panels/parallel-coordinates.
+        E.g W&B summary report `wandb.run.summary["best"]`:
+        If external EarlyStopping scheduler is used from e.g. RayTune in combination with W&B,
+        the RayTune EarlyStopping stops training function and the summary log is not sent if defined
+        after training by e.g.:
+        `wandb.run.summary["best"]=model_checkpoint.best`.
 
     Attributes
     ----------
@@ -364,6 +379,8 @@ class ModelCheckpoint(Callback):
         best metric
     best_epoch: int
         best epoch
+    best_state_dict: dict
+        best model state dictionary to restore model to its best state using trainer.model.load_state_dict(ModelCheckpoint.best_state_dict)
 
     Examples
     --------
@@ -381,13 +398,14 @@ class ModelCheckpoint(Callback):
 
     def __init__(
         self,
-        filepath: str,
+        filepath: Optional[str] = None,
         monitor: str = "val_loss",
         verbose: int = 0,
         save_best_only: bool = False,
         mode: str = "auto",
         period: int = 1,
         max_save: int = -1,
+        wb: Optional[object] = None,
     ):
         super(ModelCheckpoint, self).__init__()
 
@@ -398,18 +416,20 @@ class ModelCheckpoint(Callback):
         self.mode = mode
         self.period = period
         self.max_save = max_save
+        self.wb = wb
 
         self.epochs_since_last_save = 0
 
-        if len(self.filepath.split("/")[:-1]) == 0:
-            raise ValueError(
-                "'filepath' must be the full path to save the output weights,"
-                " including the root of the filenames. e.g. 'checkpoints/weights_out'"
-            )
+        if self.filepath:
+            if len(self.filepath.split("/")[:-1]) == 0:
+                raise ValueError(
+                    "'filepath' must be the full path to save the output weights,"
+                    " including the root of the filenames. e.g. 'checkpoints/weights_out'"
+                )
 
-        root_dir = ("/").join(self.filepath.split("/")[:-1])
-        if not os.path.exists(root_dir):
-            os.makedirs(root_dir)
+            root_dir = ("/").join(self.filepath.split("/")[:-1])
+            if not os.path.exists(root_dir):
+                os.makedirs(root_dir)
 
         if self.max_save > 0:
             self.old_files: List[str] = []
@@ -442,7 +462,8 @@ class ModelCheckpoint(Callback):
         self.epochs_since_last_save += 1
         if self.epochs_since_last_save >= self.period:
             self.epochs_since_last_save = 0
-            filepath = "{}_{}.p".format(self.filepath, epoch + 1)
+            if self.filepath:
+                filepath = "{}_{}.p".format(self.filepath, epoch + 1)
             if self.save_best_only:
                 current = logs.get(self.monitor)
                 if current is None:
@@ -454,35 +475,50 @@ class ModelCheckpoint(Callback):
                 else:
                     if self.monitor_op(current, self.best):
                         if self.verbose > 0:
-                            print(
-                                "\nEpoch %05d: %s improved from %0.5f to %0.5f,"
-                                " saving model to %s"
-                                % (
-                                    epoch + 1,
-                                    self.monitor,
-                                    self.best,
-                                    current,
-                                    filepath,
+                            if self.filepath:
+                                print(
+                                    "\nEpoch %05d: %s improved from %0.5f to %0.5f,"
+                                    " saving model to %s"
+                                    % (
+                                        epoch + 1,
+                                        self.monitor,
+                                        self.best,
+                                        current,
+                                        filepath,
+                                    )
                                 )
-                            )
+                            else:
+                                print(
+                                    "\nEpoch %05d: %s improved from %0.5f to %0.5f"
+                                    % (
+                                        epoch + 1,
+                                        self.monitor,
+                                        self.best,
+                                        current,
+                                    )
+                                )
+                        if self.wb is not None:
+                            self.wb.run.summary["best"] = current  # type: ignore[attr-defined]
                         self.best = current
                         self.best_epoch = epoch
-                        torch.save(self.model.state_dict(), filepath)
-                        if self.max_save > 0:
-                            if len(self.old_files) == self.max_save:
-                                try:
-                                    os.remove(self.old_files[0])
-                                except FileNotFoundError:
-                                    pass
-                                self.old_files = self.old_files[1:]
-                            self.old_files.append(filepath)
+                        self.best_state_dict = self.model.state_dict()
+                        if self.filepath:
+                            torch.save(self.best_state_dict, filepath)
+                            if self.max_save > 0:
+                                if len(self.old_files) == self.max_save:
+                                    try:
+                                        os.remove(self.old_files[0])
+                                    except FileNotFoundError:
+                                        pass
+                                    self.old_files = self.old_files[1:]
+                                self.old_files.append(filepath)
                     else:
                         if self.verbose > 0:
                             print(
                                 "\nEpoch %05d: %s did not improve from %0.5f"
                                 % (epoch + 1, self.monitor, self.best)
                             )
-            else:
+            if not self.save_best_only and self.filepath:
                 if self.verbose > 0:
                     print("\nEpoch %05d: saving model to %s" % (epoch + 1, filepath))
                 torch.save(self.model.state_dict(), filepath)
@@ -661,3 +697,27 @@ class EarlyStopping(Callback):
 
     def __setstate__(self, state):
         self.__dict__ = state
+
+
+class RayTuneReporter(Callback):
+    r"""Callback that allows reporting history and lr_history values to RayTune
+    during Hyperparameter tuning
+
+    Callbacks are passed as input parameters to the :obj:`Trainer` class. See
+    :class:`pytorch_widedeep.trainer.Trainer`
+
+    Examples
+    --------
+    see /examples/12_HyperParameter_tuning_w_RayTune.ipynb
+    """
+
+    def on_epoch_end(
+        self, epoch: int, logs: Optional[Dict] = None, metric: Optional[float] = None
+    ):
+        report_dict = {}
+        for k, v in self.trainer.history.items():
+            report_dict.update({k: v[-1]})
+        if hasattr(self.trainer, "lr_history"):
+            for k, v in self.trainer.lr_history.items():
+                report_dict.update({k: v[-1]})
+        tune.report(report_dict)
