@@ -13,6 +13,7 @@ from torchmetrics import Metric as TorchMetric
 from torch.utils.data import DataLoader
 from torch.optim.lr_scheduler import ReduceLROnPlateau
 
+from pytorch_widedeep.losses import ZILNLoss
 from pytorch_widedeep.metrics import Metric, MultipleMetrics
 from pytorch_widedeep.wdtypes import *  # noqa: F403
 from pytorch_widedeep.callbacks import (
@@ -80,6 +81,12 @@ class Trainer:
         - ``root_mean_squared_error``, aliases:  ``rmse``
 
         - ``root_mean_squared_log_error``, aliases: ``rmsle``
+
+        - ``zero_inflated_lognormal``, aliases: ``ziln``
+
+        - ``quantile``
+
+        - ``tweedie``
     custom_loss_function: ``nn.Module``, optional, default = None
         object of class ``nn.Module``. If none of the loss functions
         available suits the user, it is possible to pass a custom loss
@@ -683,18 +690,135 @@ class Trainer:
             If a trainer is used to predict after having trained a model, the
             ``batch_size`` needs to be defined as it will not be defined as
             the :obj:`Trainer` is instantiated
+        uncertainty: bool, default = False
+            If set to True the model activates the dropout layers and predicts
+            the each sample N times (uncertainty_granularity times) and returns
+            {max, min, mean, stdev} value for each sample
+        uncertainty_granularity: int default = 1000
+            number of times the model does prediction for each sample if uncertainty
+            is set to True
         """
-
         preds_l = self._predict(X_wide, X_tab, X_text, X_img, X_test, batch_size)
         if self.method == "regression":
             return np.vstack(preds_l).squeeze(1)
         if self.method == "binary":
             preds = np.vstack(preds_l).squeeze(1)
             return (preds > 0.5).astype("int")
+        if self.method == "qregression":
+            return np.vstack(preds_l)
         if self.method == "multiclass":
             preds = np.vstack(preds_l)
             return np.argmax(preds, 1)  # type: ignore[return-value]
 
+    def _predict_ziln(self, preds: Tensor) -> Tensor:
+        """Calculates predicted mean of zero inflated lognormal logits.
+        Adjusted implementaion of `code
+        <https://github.com/google/lifetime_value/blob/master/lifetime_value/zero_inflated_lognormal.py>`
+        Arguments:
+            preds: [batch_size, 3] tensor of logits.
+        Returns:
+            ziln_preds: [batch_size, 1] tensor of predicted mean.
+        """
+        positive_probs = torch.sigmoid(preds[..., :1])
+        loc = preds[..., 1:2]
+        scale = F.softplus(preds[..., 2:])
+        ziln_preds = positive_probs * torch.exp(loc + 0.5 * torch.square(scale))
+        return ziln_preds
+
+    def predict_uncertainty(  # type: ignore[return]
+        self,
+        X_wide: Optional[np.ndarray] = None,
+        X_tab: Optional[np.ndarray] = None,
+        X_text: Optional[np.ndarray] = None,
+        X_img: Optional[np.ndarray] = None,
+        X_test: Optional[Dict[str, np.ndarray]] = None,
+        batch_size: int = 256,
+        uncertainty_granularity=1000,
+    ) -> np.ndarray:
+        r"""Returns the predicted ucnertainty of the model for the test dataset using a
+        Monte Carlo method during which dropout layers are activated in the evaluation/prediction
+        phase and each sample is predicted N times (uncertainty_granularity times). Based on [1].
+
+        [1] Gal Y. & Ghahramani Z., 2016, Dropout as a Bayesian Approximation: Representing Model
+        Uncertainty in Deep Learning, Proceedings of the 33rd International Conference on Machine Learning
+
+        Parameters
+        ----------
+        X_wide: np.ndarray, Optional. default=None
+            Input for the ``wide`` model component.
+            See :class:`pytorch_widedeep.preprocessing.WidePreprocessor`
+        X_tab: np.ndarray, Optional. default=None
+            Input for the ``deeptabular`` model component.
+            See :class:`pytorch_widedeep.preprocessing.TabPreprocessor`
+        X_text: np.ndarray, Optional. default=None
+            Input for the ``deeptext`` model component.
+            See :class:`pytorch_widedeep.preprocessing.TextPreprocessor`
+        X_img : np.ndarray, Optional. default=None
+            Input for the ``deepimage`` model component.
+            See :class:`pytorch_widedeep.preprocessing.ImagePreprocessor`
+        X_test: Dict, Optional. default=None
+            The test dataset can also be passed in a dictionary. Keys are
+            `X_wide`, `'X_tab'`, `'X_text'`, `'X_img'` and `'target'`. Values
+            are the corresponding matrices.
+        batch_size: int, default = 256
+            If a trainer is used to predict after having trained a model, the
+            ``batch_size`` needs to be defined as it will not be defined as
+            the :obj:`Trainer` is instantiated
+        uncertainty_granularity: int default = 1000
+            number of times the model does prediction for each sample if uncertainty
+            is set to True
+
+        Returns
+        -------
+            method == regression : np.ndarray
+                {max, min, mean, stdev} values for each sample
+            method == binary : np.ndarray
+                {mean_cls_0_prob, mean_cls_1_prob, predicted_cls} values for each sample
+            method == multiclass : np.ndarray
+                {mean_cls_0_prob, mean_cls_1_prob, mean_cls_2_prob, ... , predicted_cls} values for each sample
+
+        """
+        preds_l = self._predict(
+            X_wide,
+            X_tab,
+            X_text,
+            X_img,
+            X_test,
+            batch_size,
+            uncertainty_granularity,
+            uncertainty=True,
+        )
+        preds = np.vstack(preds_l)
+        samples_num = int(preds.shape[0] / uncertainty_granularity)
+        if self.method == "regression":
+            preds = preds.squeeze(1)
+            preds = preds.reshape((uncertainty_granularity, samples_num))
+            return np.array(
+                (
+                    preds.max(axis=0),
+                    preds.min(axis=0),
+                    preds.mean(axis=0),
+                    preds.std(axis=0),
+                )
+            ).T
+        if self.method == "qregression":
+            raise ValueError(
+                "Currently predict_uncertainty is not supported for qregression method"
+            )
+        if self.method == "binary":
+            preds = preds.squeeze(1)
+            preds = preds.reshape((uncertainty_granularity, samples_num))
+            preds = preds.mean(axis=0)
+            probs = np.zeros([preds.shape[0], 3])
+            probs[:, 0] = 1 - preds
+            probs[:, 1] = preds
+            return probs
+        if self.method == "multiclass":
+            preds = preds.reshape(uncertainty_granularity, samples_num, preds.shape[1])
+            preds = preds.mean(axis=0)
+            preds = np.hstack((preds, np.vstack(np.argmax(preds, 1))))
+            return preds
+
     def predict_proba(  # type: ignore[return]
         self,
         X_wide: Optional[np.ndarray] = None,
@@ -942,14 +1066,11 @@ class Trainer:
             for callback in self.callback_container.callbacks:
                 if callback.__class__.__name__ == "ModelCheckpoint":
                     if callback.save_best_only:
-                        filepath = "{}_{}.p".format(
-                            callback.filepath, callback.best_epoch + 1
-                        )
                         if self.verbose:
                             print(
                                 f"Model weights restored to best epoch: {callback.best_epoch + 1}"
                             )
-                        self.model.load_state_dict(torch.load(filepath))
+                        self.model.load_state_dict(callback.best_state_dict)
                     else:
                         if self.verbose:
                             print(
@@ -1032,7 +1153,11 @@ class Trainer:
     def _train_step(self, data: Dict[str, Tensor], target: Tensor, batch_idx: int):
         self.model.train()
         X = {k: v.cuda() for k, v in data.items()} if use_cuda else data
-        y = target.view(-1, 1).float() if self.method != "multiclass" else target
+        y = (
+            target.view(-1, 1).float()
+            if self.method not in ["multiclass", "qregression"]
+            else target
+        )
         y = y.to(device)
 
         self.optimizer.zero_grad()
@@ -1043,6 +1168,7 @@ class Trainer:
         else:
             loss = self.loss_fn(y_pred, y)
             score = self._get_score(y_pred, y)
+        # TODO raise exception if the loss is exploding with non scaled target values
         loss.backward()
         self.optimizer.step()
 
@@ -1056,7 +1182,11 @@ class Trainer:
         self.model.eval()
         with torch.no_grad():
             X = {k: v.cuda() for k, v in data.items()} if use_cuda else data
-            y = target.view(-1, 1).float() if self.method != "multiclass" else target
+            y = (
+                target.view(-1, 1).float()
+                if self.method not in ["multiclass", "qregression"]
+                else target
+            )
             y = y.to(device)
 
             y_pred = self.model(X)
@@ -1078,6 +1208,8 @@ class Trainer:
                 score = self.metric(y_pred, y)
             if self.method == "binary":
                 score = self.metric(torch.sigmoid(y_pred), y)
+            if self.method == "qregression":
+                score = self.metric(y_pred, y)
             if self.method == "multiclass":
                 score = self.metric(F.softmax(y_pred, dim=1), y)
             return score
@@ -1102,7 +1234,7 @@ class Trainer:
             k: v for k, v in zip(tabnet_backbone.column_idx.keys(), feat_imp)  # type: ignore[operator, union-attr]
         }
 
-    def _predict(
+    def _predict(  # noqa: C901
         self,
         X_wide: Optional[np.ndarray] = None,
         X_tab: Optional[np.ndarray] = None,
@@ -1110,6 +1242,9 @@ class Trainer:
         X_img: Optional[np.ndarray] = None,
         X_test: Optional[Dict[str, np.ndarray]] = None,
         batch_size: int = 256,
+        uncertainty_granularity=1000,
+        uncertainty: bool = False,
+        quantiles: bool = False,
     ) -> List:
         r"""Private method to avoid code repetition in predict and
         predict_proba. For parameter information, please, see the .predict()
@@ -1142,20 +1277,45 @@ class Trainer:
 
         self.model.eval()
         preds_l = []
+
+        if uncertainty:
+            for m in self.model.modules():
+                if m.__class__.__name__.startswith("Dropout"):
+                    m.train()
+            prediction_iters = uncertainty_granularity
+        else:
+            prediction_iters = 1
+
         with torch.no_grad():
-            with trange(test_steps, disable=self.verbose != 1) as t:
-                for i, data in zip(t, test_loader):
-                    t.set_description("predict")
-                    X = {k: v.cuda() for k, v in data.items()} if use_cuda else data
-                    preds = (
-                        self.model(X) if not self.model.is_tabnet else self.model(X)[0]
-                    )
-                    if self.method == "binary":
-                        preds = torch.sigmoid(preds)
-                    if self.method == "multiclass":
-                        preds = F.softmax(preds, dim=1)
-                    preds = preds.cpu().data.numpy()
-                    preds_l.append(preds)
+            with trange(uncertainty_granularity, disable=uncertainty is False) as t:
+                for i, k in zip(t, range(prediction_iters)):
+                    t.set_description("predict_UncertaintyIter")
+
+                    with trange(
+                        test_steps, disable=self.verbose != 1 or uncertainty is True
+                    ) as tt:
+                        for j, data in zip(tt, test_loader):
+                            tt.set_description("predict")
+                            X = (
+                                {k: v.cuda() for k, v in data.items()}
+                                if use_cuda
+                                else data
+                            )
+                            preds = (
+                                self.model(X)
+                                if not self.model.is_tabnet
+                                else self.model(X)[0]
+                            )
+                            if self.method == "binary":
+                                preds = torch.sigmoid(preds)
+                            if self.method == "multiclass":
+                                preds = F.softmax(preds, dim=1)
+                            if self.method == "regression" and isinstance(
+                                self.loss_fn, ZILNLoss
+                            ):
+                                preds = self._predict_ziln(preds)
+                            preds = preds.cpu().data.numpy()
+                            preds_l.append(preds)
         self.model.train()
         return preds_l
 
@@ -1164,7 +1324,10 @@ class Trainer:
             class_weight = torch.tensor(class_weight).to(device)
         if custom_loss_function is not None:
             return custom_loss_function
-        elif self.method != "regression" and "focal_loss" not in objective:
+        elif (
+            self.method not in ["regression", "qregression"]
+            and "focal_loss" not in objective
+        ):
             return alias_to_loss(objective, weight=class_weight)
         elif "focal_loss" in objective:
             return alias_to_loss(objective, alpha=alpha, gamma=gamma)
